/*****************************************************************************
 *                                                                           *
 *  Copyright (c) 1993, 1994, Elan Feingold (elan@tasha.cheme.cornell.edu)   *
 *                                                                           *
 *     PERMISSION TO USE, COPY, MODIFY, AND TO DISTRIBUTE THIS SOFTWARE      *
 *     AND ITS DOCUMENTATION FOR ANY PURPOSE IS HEREBY GRANTED WITHOUT       *
 *     FEE, PROVIDED THAT THE ABOVE COPYRIGHT NOTICE APPEAR IN ALL           *
 *     COPIES AND MODIFIED COPIES AND THAT BOTH THAT COPYRIGHT NOTICE AND    *
 *     THIS PERMISSION NOTICE APPEAR IN SUPPORTING DOCUMENTATION.  THERE     *
 *     IS NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR      *
 *     ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS       *
 *     OR IMPLIED WARRANTY.                                                  *
 *                                                                           *
 *****************************************************************************/

#include "factory.hh"
#include "bitmaps.hh"
#include "callbacks.hh"
#include "x.hh"

/************************************************************************ 
 *  FUNCTION: CB_ConveyorBelt
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     This is the callback for drawing conveyor belts.  If it is a LIFE 
 *   callback, then just draw it, otherwise, if it a STEP callback, and 
 *   it's the first time around for this particular conveyor belt, fork
 *   off a thread that is responsible for animating the belt at the same
 *   speed that object pass over the belt.
 *  NOTES: 
 ************************************************************************/
void CB_ConveyorBelt(CallData *pData)
{
  static Boolean fSpawned[MAX_STAGES];
  Int            j;

  /* Fork off an artist thread if needed */
  if (pData->iType != CALLBACK_LIFE && fSpawned[pData->iStage] == FALSE)
    {
      t_forkRT((ThreadFunc)CB_DrawConveyorBelt, (any_ptr)pData, "Belt Artist", 
	       PRIORITY_LOW);
      fSpawned[pData->iStage] = TRUE;
    }
  else
    /* Draw it.  Must loop b/c might be superscalar conveyor belt */
    for (j=0; j!=pData->iWidth; j++)
      DrawBitmap(BITMAP(cbelt1), 
		 pData->iObjX, pData->iObjY + j * (cbelt1_height+55));
}


/************************************************************************ 
 *  FUNCTION: CB_DrawConveyorBelt
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     This is the artist thread for conveyor belts.
 *  NOTES: 
 ************************************************************************/
void CB_DrawConveyorBelt(any_ptr vData)
{
  CallData   *pData = (CallData *)vData;
  Int         x=pData->iObjX, y=pData->iObjY, i, j, iWidth=pData->iWidth;
  PCondition  pCond = cond_create();
  
  cond_registerRT(pCond, pData->iStepTime/2);

  /* Cycle every time a blob moves a step (approximately) */
  for(i=0; ;i++)
    {
      if(i % 2)
	for (j=0; j!=iWidth; j++)
	  DrawBitmap(BITMAP(cbelt1), x, y+j*(cbelt1_height+55));
      else
	for (j=0; j!=iWidth; j++)
	  DrawBitmap(BITMAP(cbelt2), x, y+j*(cbelt1_height+55));

      t_wait(pCond);
    }
}


/************************************************************************ 
 *  FUNCTION: ConveyorBeltCoordServer
 *  HISTORY: 
 *     11/23/93  ESF  Created 
 *  PURPOSE: 
 *     Based on the CallData, returns the coordinates at which the blob
 *   should be painted.  Takes into accound the step at which the blob
 *   is at, the shape of the blob, and the belt that the blob is on
 *   (for superscalar belts).
 *  NOTES: 
 ************************************************************************/
void CB_ConveyorBeltCoordServer(CallData *pData)
{
  Int iObjHeight;

  /* Hum, what is the height of the object on the belt? */
  if (pData->iStage == 0)      /* Flour */
    iObjHeight = flour_height;
  else if (pData->iStage == 2) /* Dough */
    iObjHeight = dough_height;
  else if (pData->iStage = 4)  /* Bread */
    iObjHeight = bread_height;

  /* Return a set of coords, based on step and which belt blob is on */
  pData->iDrawX = pData->iObjX + pData->iStep*80;
  pData->iDrawY = pData->iObjY - iObjHeight + 
    pData->iPipeOwned*(cbelt1_height+55);
}


/************************************************************************ 
 *  FUNCTION: CB_Mixer
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     The Mixer callback.  Handles animation for the mixer.
 *  NOTES: 
 ************************************************************************/
void CB_Mixer(CallData *pData)
{
  char *data;
  int   width, height;

  switch (pData->iClick % 10)
    {
    case 0:
      data=wash_bits, width=wash_width, height=wash_height;
      break;
    case 1:
      data=wash0_bits, width=wash0_width, height=wash0_height;
      break;
    case 2:
      data=wash1_bits, width=wash1_width, height=wash1_height;
      break;
    case 3:
      data=wash2_bits, width=wash2_width, height=wash2_height;
      break;
    case 4:
      data=wash3_bits, width=wash3_width, height=wash3_height;
      break;
    case 5:
      data=wash4_bits, width=wash4_width, height=wash4_height;
      break;
    case 6:
      data=wash5_bits, width=wash5_width, height=wash5_height;
      break;
    case 7:
      data=wash6_bits, width=wash6_width, height=wash6_height;
      break;
    case 8:
      data=wash7_bits, width=wash7_width, height=wash7_height;
      break;
    case 9:
      data=wash0_bits, width=wash0_width, height=wash0_height;
      break;
    default:
      data=NULL;
    }

  if (data)
    DrawBitmap(data, width, height, pData->iObjX, pData->iObjY);
}


/************************************************************************ 
 *  FUNCTION: CB_Oven
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     The callback for the oven.  Handles animation for the oven.
 *  NOTES: 
 ************************************************************************/
void CB_Oven(CallData *pData)
{
  char *data;
  int   width, height;

  switch (pData->iClick % 3)
    {
    case 0:
      data=oven0_bits, width=oven0_width, height=oven0_height;
      break;
    case 1:
      data=oven1_bits, width=oven1_width, height=oven1_height;
      break;
    case 2:
      data=oven2_bits, width=oven2_width, height=oven2_height;
      break;
    default:
      data=NULL;
    }

  if (data)
    DrawBitmap(data, width, height, pData->iObjX, pData->iObjY);
}


/************************************************************************ 
 *  FUNCTION: CB_Truck
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     Callback for the truck stage.  Handles drawing it, and also is
 *   a handler for CALLBACK_FINISHED, which should only be given to the
 *   last stage in a factory.  In this factory, the truck drives away 
 *   when this callback occurs, and then drives back.
 *  NOTES: 
 ************************************************************************/
void CB_Truck(CallData *pData)
{
  if (pData->iType == CALLBACK_LIFE)
    DrawBitmap(BITMAP(truck), pData->iObjX, pData->iObjY);
  else if (pData->iType == CALLBACK_FINISHED)
    {
      Int          i;
      PCondition   pCond = cond_create();

      cond_registerRT(pCond, Millisec(90000));
      
      /* Drive truck away */
      for(i=1; i!=12; i++)
	{
	  EraseBitmap(BITMAP(truck), pData->iObjX + (i-1)*20, pData->iObjY); 
	  DrawBitmap(BITMAP(truck), pData->iObjX + i*20, pData->iObjY);
	  t_wait(pCond);
	}

      /* Bring truck back */
      for(i=10; i!=-1; i--)
	{
	  EraseBitmap(BITMAP(truck), pData->iObjX + (i+1)*20, pData->iObjY); 
	  DrawBitmap(BITMAP(truck), pData->iObjX + i*20, pData->iObjY);
	  t_wait(pCond);
	}

      cond_unregisterRT(pCond);
      cond_destroy(pCond);
    }
}


/************************************************************************ 
 *  FUNCTION: CB_Blob
 *  HISTORY: 
 *     11/23/93  ESF  Created 
 *  PURPOSE: 
 *     Handles drawing for blobs.  Based on whih stage they're in, draws
 *   then appropriately.
 *  NOTES: 
 ************************************************************************/
void CB_Blob(CallData *pData)
{
  char *data;
  int   width, height;
  
  switch (pData->iStage)
    {
    case 0:  /* First Conveyor */
      data=flour_bits, width=flour_width, height=flour_height;
      break;
    case 2:  /* Second Conveyor */
      data=dough_bits, width=dough_width, height=dough_height;
      break;
    case 4:  /* Third Conveyor */
      data=bread_bits, width=bread_width, height=bread_height;
      break;
    default:
      data=NULL;
    }

  /* either draw or erase, depending on the callback */
  if (data)
    {
      if (pData->iType == CALLBACK_ACQUIRE)
	DrawBitmap(data, width, height, pData->iDrawX, pData->iDrawY);
      else if (pData->iType == CALLBACK_RELEASE)
	EraseBitmap(data, width, height, pData->iDrawX, pData->iDrawY);
    }
}
