/*****************************************************************************
 *                                                                           *
 *  Copyright (c) 1993, 1994, Elan Feingold (elan@tasha.cheme.cornell.edu)   *
 *                                                                           *
 *     PERMISSION TO USE, COPY, MODIFY, AND TO DISTRIBUTE THIS SOFTWARE      *
 *     AND ITS DOCUMENTATION FOR ANY PURPOSE IS HEREBY GRANTED WITHOUT       *
 *     FEE, PROVIDED THAT THE ABOVE COPYRIGHT NOTICE APPEAR IN ALL           *
 *     COPIES AND MODIFIED COPIES AND THAT BOTH THAT COPYRIGHT NOTICE AND    *
 *     THIS PERMISSION NOTICE APPEAR IN SUPPORTING DOCUMENTATION.  THERE     *
 *     IS NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR      *
 *     ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS       *
 *     OR IMPLIED WARRANTY.                                                  *
 *                                                                           *
 *****************************************************************************/

#include "factory.hh"
#include "x.hh"

#include "thread.h"
#include "debug.h"

/* Some handy macros */
#define PipeEntry(x, y) (x+y*pStage[iStage].iWidth)
#define BuildCallData(callData, iCallType, iClk, iPipe)   \
  callData.iType      = iCallType; \
  callData.iObjX      = pStage[iStage].iObjX; \
  callData.iObjY      = pStage[iStage].iObjY; \
  callData.iStage     = iStage; \
  callData.iStepTime  = pStage[iStage].iStepTime; \
  callData.iStep      = iStep; \
  callData.iWidth     = pStage[iStage].iWidth; \
  callData.iPipeOwned = iPipe; \
  callData.iClick     = iClk
  
/* Local prototypes */
void FactoryStart(any_ptr foo);
void MainLoop(any_ptr foo);


/************************************************************************ 
 *  FUNCTION: Factory::Factory
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *     11/24/93  ESF  Added callbacks
 *  PURPOSE: 
 *    Initializes the factory, as well as all of the stages.
 *  NOTES: 
 ************************************************************************/
Factory::Factory(Int iWidth, Int iHeight)
{
  /* Init. variables */
  iNumStages    = 0;
  this->iWidth  = iWidth;
  this->iHeight = iHeight;

  /* Init callbacks */
  for (Int i=0; i!=MAX_STAGES; i++)
    {
      pStage[i].LifeCallBack = NULL;
      pStage[i].ClickCallBack = NULL;
      pStage[i].AcquireCallBack = NULL;
      pStage[i].ReleaseCallBack = NULL;
      pStage[i].FinalCallBack = NULL;
      pStage[i].CoordServerCallBack = NULL;
    }
}


/************************************************************************ 
 *  FUNCTION: Factory::StageAdd
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *  PURPOSE: 
 *     Adds a stage to the factory.  strName is the name of the stage, 
 *   used for debugging purposes.  iPipeSteps is the number of steps in
 *   the stage, which is exactly the same as pipeline steps in a piepline.
 *   Conveyor belts have multiple steps.  iClicks is the granularity of a
 *   step.  With some stages (i.e. mixer) where there is only one step,
 *   per se, but there is more than one thing happening (animations).
 *   Stages can register a CALLBACK_CLICK and create these animations, 
 *   seting iClicks to the number of frames in the animation.  iStepTime
 *   is the time it takes per step in the stage.  If there is more than
 *   one click/step, then this time will get broken down and the callback
 *   will get called multiple times per step.
 *  NOTES: 
 ************************************************************************/
Int Factory::StageAdd(String strName, Int iWidth, Int iPipeSteps, Int iClicks,
		      Int iStepTime)
{
  Int i;

  /* I don't think so :) */
  if (iNumStages==MAX_STAGES)
    return -1;

  /* Build the stage */
  pStage[iNumStages].strName        = strName;
  pStage[iNumStages].iPipeSteps     = iPipeSteps;
  pStage[iNumStages].iWidth         = iWidth;
  pStage[iNumStages].iClicksPerStep = iClicks;
  pStage[iNumStages].pStepMutex     = (Mutex *)
    MEM_Alloc(sizeof(Mutex)*iPipeSteps*iWidth);
  pStage[iNumStages].pStepCond      = (PCondition *)
    MEM_Alloc(sizeof(PCondition)*iPipeSteps);
  pStage[iNumStages].iStepTime      = iStepTime;

  /* Initialize the Mutexes */
  for (i=0; i!=iPipeSteps*iWidth; i++)
    pStage[iNumStages].pStepMutex[i] = 0;

  /* Create the Conditions */
  for (i=0; i!=iPipeSteps; i++)
    pStage[iNumStages].pStepCond[i]  = (PCondition)cond_create();

  /* The final condition, which blobs wait on when they have finished. */
  pCondFinal = cond_create();
  
  return(iNumStages++);
}


/************************************************************************ 
 *  FUNCTION: Factory::~Factory
 *  HISTORY: 
 *
 *  PURPOSE: 
 *     To free up memory taken by factory.
 *  NOTES: 
 *     MUST IMPLEMENT EVENTUALLY.  Right now not really important, since
 *   end of factory ==> end of program.
 ************************************************************************/
Factory::~Factory()
{
  /* Twiddle thumbs */
}


/************************************************************************ 
 *  FUNCTION: Factory::StageRegisterCoords
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     Lets the factory know where the stage is located in its world space.
 *  NOTES: 
 ************************************************************************/
void Factory::StageRegisterCoords(Int iStage, Int iObjX, Int iObjY)
{
  pStage[iStage].iObjX = iObjX;
  pStage[iStage].iObjY = iObjY;
}


/************************************************************************ 
 *  FUNCTION: Factory::StageRegisterCallback 
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *     11/24/93  ESF  Added callbacks
 *  PURPOSE: 
 *     Used to register callbacks.  Can register multiple callbacks at
 *   the same time, logcally OR-ing them together.
 *  NOTES: 
 ************************************************************************/
void Factory::StageRegisterCallback(Int iStage, Int iType, 
				    void (*fcn)(CallData *))
{
  if (iType & CALLBACK_LIFE)
    pStage[iStage].LifeCallBack = fcn;
  
  if (iType & CALLBACK_ACQUIRE)
    pStage[iStage].AcquireCallBack = fcn;

  if (iType & CALLBACK_RELEASE)
    pStage[iStage].ReleaseCallBack = fcn;

  if (iType & CALLBACK_CLICK)
    pStage[iStage].ClickCallBack = fcn;

  if (iType & CALLBACK_FINISHED)
    pStage[iStage].FinalCallBack = fcn;

  if (iType & CALLBACK_COORDSERVER)
    pStage[iStage].CoordServerCallBack = fcn;
}


/************************************************************************ 
 *  FUNCTION: Factory::AcquireResource
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *     11/21/93  ESF  Changed to work with superscalar pipelines
 *  PURPOSE: 
 *     Acquires the desired step in the desired piepline in the desired 
 *   stage.  Doesn't return until the stage is in the caller blob's 
 *   possesion.
 *  NOTES: 
 ************************************************************************/
Int Factory::AcquireResource(Int iStage, Int iStep, Int iOwned)
{
  Int       i;
  CallData  callData;

 GetRes:
  t_cs_entry();

  /* If this is the first step of a stage, doesn't own a specific pipe */
  if (iStep==0)
    iOwned=-1;

  /* If the step or stage goes beyond that allowed, ignore and return */
  if (iStage >= iNumStages ||
      iStep  >= QuerySteps(iStage))
    return -1;
  
  /* If the resource is not owned, own it  */
  for (i=0; i!=pStage[iStage].iWidth; i++)
    if (!pStage[iStage].pStepMutex[PipeEntry(i, iStep)] &&
	(iOwned==-1 || i==iOwned))
      {
	pStage[iStage].pStepMutex[PipeEntry(i, iStep)] = 1;

	/* Callback stuff */
	BuildCallData(callData, CALLBACK_ACQUIRE, 0, i);
	
	/* If the stage has a coordinate server for the object, call it 
	 * to obtain coordinate information so that other callbacks will
	 * know how to draw the blob.
	 */
	
	if (pStage[iStage].CoordServerCallBack)
	  pStage[iStage].CoordServerCallBack(&callData);
	
	/* If the stage has an Acquire Callback, call it */
	if (pStage[iStage].AcquireCallBack)
	  pStage[iStage].AcquireCallBack(&callData);	

	t_cs_exit();
	
	return(i);
      }
  
  /* We have to wait for it */
  t_cs_exit();
  t_wait(pStage[iStage].pStepCond[iStep]);

  /* At this point we have to go back and fight for it again */
  goto GetRes;
}


/************************************************************************ 
 *  FUNCTION: Factory::ReleaseResource
 *  HISTORY: 
 *     11/16/93  ESF  Created
 *     11/21/93  ESF  Fixed dumb bug
 *     11/21/93  ESF  Changed to work with superscalar pipelines
 *  PURPOSE: 
 *     Releases a resource.  Checks to see if we are in the last
 *   stage of the factory, which is assumed to be a collection stage
 *   (i.e. width>1, step==1).  If so, makes the blobs wait, until
 *   the collection stage has filled, and then calls CALLBACK_FINAL
 *   callback if one is registered.
 *  NOTES: 
 ************************************************************************/
void Factory::ReleaseResource(Int iStage, Int iStep, Int iOwned)
{
  CallData callData;

  t_cs_entry();

  /* Are we the last blob to finish the factory path? */
  if (iStage==iNumStages-1)
    {
      Boolean fFinished = TRUE;
      
      /* Is everyone waiting? */
      for (Int k=0; k!=pStage[iStage].iWidth; k++)
	if (!pStage[iStage].pStepMutex[k])
	  fFinished = FALSE;
      
      /* If so, then Deal with every blob in the last stage */
      if (fFinished)
	{
	  CallData callData;
	  
	  if (pStage[iStage].FinalCallBack)
	    {
	      BuildCallData(callData, CALLBACK_FINISHED, 0, 0);
	      pStage[iStage].FinalCallBack(&callData);
	    }
	  
	  /* Let all of the blobs that are waiting on this finish */
	  while (!cond_is_empty(pCondFinal))
	    t_sig(pCondFinal, NULL, TRUE);
	  
	  /* For this blob to find in a jiffy */
	  t_sig(pCondFinal, NULL, TRUE);
	}

      /* Wait around */
      t_wait(pCondFinal);	
    }

  /* Resource Release callback */
  BuildCallData(callData, CALLBACK_RELEASE, 0, iOwned);
  
  /* If the stage has a coordinate server for the object, call it 
   * to obtain coordinate information so that other callbacks will
   * know how to draw the blob.
   */
  
  if (pStage[iStage].CoordServerCallBack)
    pStage[iStage].CoordServerCallBack(&callData);

  /* If the stage has a Release Callback, call it */
  if (pStage[iStage].ReleaseCallBack)
    pStage[iStage].ReleaseCallBack(&callData);

  /* Free the resource and signal anyone that's waiting on it */
  pStage[iStage].pStepMutex[PipeEntry(iOwned, iStep)] = 0;
  t_sig(pStage[iStage].pStepCond[iStep], (any_ptr)1, FALSE);

  t_cs_exit();
}


/************************************************************************ 
 *  FUNCTION: Factory::UseResource
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *  PURPOSE: 
 *     Simulates using the resource.  Waits around the appropriate time,
 *   calling click callbacks as needed.
 *  NOTES: 
 ************************************************************************/
void Factory::UseResource(Int iStage, Int iStep, Int iPipeOwned)
{
  Int         i;
  PCondition  pCond = cond_create();
  CallData    callData;
  
  /* If we are supposed to wait at all */
  if (pStage[iStage].iStepTime)
    {
      /* Want to spring into action every click */
      cond_registerRT(pCond, 
		      pStage[iStage].iStepTime/pStage[iStage].iClicksPerStep);
      
      for (i=0; i!=pStage[iStage].iClicksPerStep; i++)
	{
	  /* Resource Click callback */
	  BuildCallData(callData, CALLBACK_CLICK, i, iPipeOwned);
	  
	  if (pStage[iStage].ClickCallBack)
	    pStage[iStage].ClickCallBack(&callData);
	  
	  t_wait(pCond);
	} 
      cond_unregisterRT(pCond);
    }
  else
    ; /* Free usage */

  cond_destroy(pCond);
}


/************************************************************************ 
 *  FUNCTION: Factory::QueryStages
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *  PURPOSE: 
 *     Speaks for itself.
 *  NOTES: 
 ************************************************************************/
UInt Factory::QueryStages(void)
{
  return (iNumStages);
}


/************************************************************************ 
 *  FUNCTION: Factory::QuerySteps 
 *  HISTORY: 
 *     11/16/93  ESF  Created 
 *  PURPOSE: 
 *     Nuff Said.
 *  NOTES: 
 ************************************************************************/
UInt Factory::QuerySteps(Int iStage)
{
  return (pStage[iStage].iPipeSteps);
}


/************************************************************************ 
 *  FUNCTION: Factory::QueryStageName
 *  HISTORY: 
 *     11/17/93  ESF  Created 
 *  PURPOSE: 
 *     Need I say more?
 *  NOTES: 
 ************************************************************************/
String Factory::QueryStageName(Int iStage)
{
  return (pStage[iStage].strName);
}


/************************************************************************ 
 *  FUNCTION: Factory::Start
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     Creates the Window, calls all of the object's CALLBACK_LIFE
 *   callbacks, if they've registered them.
 *  NOTES: 
 ************************************************************************/
void Factory::Start(int argc, char **argv)
{
  CallData  callData;
  Int       i;

  CreateXWindow("The Factory", 640, 480, argc, argv); 

  callData.iStep  = 0;
  callData.iClick = 0;
  
  /* Setup the Life callbacks */
  for (i=0; i!=iNumStages; i++)
    {
      callData.iWidth = pStage[i].iWidth;
      callData.iObjX  = pStage[i].iObjX;
      callData.iObjY  = pStage[i].iObjY;
      callData.iType  = CALLBACK_LIFE;
        	
      if (pStage[i].LifeCallBack)
	pStage[i].LifeCallBack(&callData);
    }

  /* Get the threads rolling */
  t_debug(fopen("log", "w")); 
  t_startRT((ThreadFunc)FactoryStart, (any_ptr)SpawnFunc, "Factory", 
	    PRIORITY_HIGH, FALSE);
}


/************************************************************************ 
 *  FUNCTION: Factory::Start
 *  HISTORY: 
 *     11/22/93  ESF  Created 
 *  PURPOSE: 
 *     Here is where the threads package wakes up.  Spawns off the X
 *   main loop, and then calls the registered spawning function if
 *   there is one registered.
 *  NOTES: 
 ************************************************************************/
void FactoryStart(any_ptr SpawnFunc)
{
  t_forkRT((ThreadFunc)MainLoop, NULL, "X Main Loop", PRIORITY_LOW);
  if (SpawnFunc)
    t_forkRT((ThreadFunc)SpawnFunc, NULL, "Factory Spawner", PRIORITY_LOW);
}
