#ifndef _FACTORY
#define _FACTORY

extern "C" {
#include "thread/standard.h"
#include "thread/types.h"
#include "thread/thread.h"
}

#ifdef LINUX_GCC
#define Millisec(x) ((x)==0 ? 0 : (x/10000))
#elif SUN_GCC
#define Millisec(x) (x)
#endif

#define MAX_STAGES 32

#define CALLBACK_LIFE          1
#define CALLBACK_FINISHED      2
#define CALLBACK_CLICK         4
#define CALLBACK_ACQUIRE      16
#define CALLBACK_RELEASE      32
#define CALLBACK_COORDSERVER  64

typedef Boolean Mutex;

#define NextStep(factory, stage, step) \
  (step < factory.QuerySteps(stage)-1 ? step+1 : 0)  
#define NextStage(factory, stage, step) \
  (step < factory.QuerySteps(stage)-1 ? stage : stage+1)


/* A factory is built of stages.  Each state represents an interval in which
 * a Blob is manipulated in some way: either transformed (i.e. dough baked)
 * or translated (i.e. moved down a conveyor belt).  Each stage has three
 * parameters:  
 *     
 *   o width: The number of items that can be serviced at the same time
 *            by a stage.  If the width is > 1, then the model is 
 *            assumed to operate identically to a superscalar pipeline,
 *            but without restrictions as far as issuing Blobs.
 *   o steps: The number of steps in the stage (like a pipeline).
 *   o clicks_per_step: The number of times the CLICK callback gets called
 *            for each step.  For example, in an OVEN, steps = 1, but we 
 *            want to represent different stages in the baking, so we could 
 *            set clicks_per_step to 3, for example.
 */


/* Callback data */
typedef struct _CallData
{
  Int  iStage, iStep, iPipeOwned;
  Int  iObjX, iObjY;
  Int  iType, iClick, iWidth, iStepTime;
  Int  iDrawX, iDrawY;
} CallData;


/* See factory.C for explanations */
class Factory
{
 public:
  
  Factory(Int iWidth, Int iHeight);
  ~Factory();
  
  /* Misc methods */
  void  Start(int argc, char **argv);
  void  RegisterSpawnFunction(ThreadFunc fcn) { SpawnFunc = fcn; };
  
  /* Factory Stage Methods */
  Int   StageAdd(String strName, Int iWidth, Int iPipeSteps, 
		 Int iClicksPerStep, Int iStepTime);
  void  StageRegisterCallback(Int iStage, Int iType, void (*fcn)(CallData *));
  void  StageRegisterObjectCallback(Int iStage, Int iType, 
				    void (*fcn)(CallData *));
  void  StageRegisterCoords(Int iStage, Int iObjX, Int iObjY);

  /* Factory Resource Methods */
  Int   AcquireResource(Int iStage, Int iStep, Int iPipeOwned);
  void  UseResource(Int iStage, Int iStep, Int iPipeOwned);
  void  ReleaseResource(Int iStage, Int iStep, Int iPipeOwned);

  /* Query methods */
  UInt   QueryStages(void);
  UInt   QuerySteps(Int iStage);
  String QueryStageName(Int iStage);

 private:
  
  typedef struct _Stage
    {
      String       strName;
      Int          iPipeSteps, iWidth, iClicksPerStep, iStepTime;
      Int          iObjX, iObjY;
      Mutex       *pStepMutex;
      PCondition  *pStepCond;

      void       (*LifeCallBack)(CallData *);   
      void       (*FinalCallBack)(CallData *);   
      void       (*AcquireCallBack)(CallData *);
      void       (*ReleaseCallBack)(CallData *);
      void       (*ClickCallBack)(CallData *);  
      void       (*CoordServerCallBack)(CallData *);  
    } Stage;

  Stage         pStage[MAX_STAGES];
  Int           iNumStages, iWidth, iHeight;
  ThreadFunc    SpawnFunc;
  PCondition    pCondFinal; 
};

#endif



