/*****************************************************************************
 *                                                                           *
 *  Copyright (c) 1993, 1994, Elan Feingold (elan@tasha.cheme.cornell.edu)   *
 *                                                                           *
 *     PERMISSION TO USE, COPY, MODIFY, AND TO DISTRIBUTE THIS SOFTWARE      *
 *     AND ITS DOCUMENTATION FOR ANY PURPOSE IS HEREBY GRANTED WITHOUT       *
 *     FEE, PROVIDED THAT THE ABOVE COPYRIGHT NOTICE APPEAR IN ALL           *
 *     COPIES AND MODIFIED COPIES AND THAT BOTH THAT COPYRIGHT NOTICE AND    *
 *     THIS PERMISSION NOTICE APPEAR IN SUPPORTING DOCUMENTATION.  THERE     *
 *     IS NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR      *
 *     ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS       *
 *     OR IMPLIED WARRANTY.                                                  *
 *                                                                           *
 *****************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>

#include <X11/bitmaps/Up>
#include <X11/bitmaps/Right>
#include <X11/bitmaps/Down>
#include <X11/bitmaps/Left>

#include "types.h"
#include "x.hh"

#define _hack_ticks_ 50


/* Local globals */
Widget         Toplevel, Box;
XtAppContext   context;
Pixmap         pixmap;
GC             gcDraw, gcErase;
XtIntervalId   timer;

/* Stefan's Hash table */
typedef struct bithash
{
  char *id;
  Pixmap xid;
  struct bithash *next;
} bithash;

#define TABLE_SIZE 101
bithash *bit_table[TABLE_SIZE];

/* Local utility functions */
bithash *hash_find_bitmap (char *bitmap);
void hash_bitmap (char *id, Pixmap xid);

/* Local callbacks */
void RefreshMe(Widget w, XtPointer closure, XtPointer calldata);
void ExitNow(Widget w, XtPointer closure, XtPointer calldata);
void TickleXWindow(Widget w, XtPointer closure, XtPointer calldata);


/************************************************************************ 
 *  FUNCTION: CreateXWindow
 *  HISTORY: 
 *     ??/??/93  ???  Created
 *     11/??/93   SM  Added hash table, to speed up graphics.
 *     11/22/93  ESF  Changed to work with Xt, event handlers, etc.
 *  PURPOSE: 
 *     Creates a window of size (iWidth x iHeight).  Then it adds an
 *   app timer (to be explained later), and two event handlers, for
 *   expose events and for mouse clicks.  It then creates two graphics
 *   contexts, a backing store pixmap, and finally realizes the top
 *   level widget.
 *  NOTES: 
 ************************************************************************/
void CreateXWindow(char *szName, Int iWidth, Int iHeight, int argc, 
		   char **argv)
{
  Arg args[10];

  /* Create the application's main window */
  Toplevel = XtAppInitialize(&context, "Factory", NULL, 0, &argc, argv, 
			       NULL, NULL, 0);

  /* Create the widget, and resize it appropriately */
  XtSetArg(args[0], XtNwidth, iWidth);
  XtSetArg(args[1], XtNheight, iHeight);
  Box = XtCreateManagedWidget("box", boxWidgetClass, Toplevel, args, 2);

  /* Create an app timeout to serve as a tickler */
  timer = XtAppAddTimeOut(context, _hack_ticks_, TickleXWindow, Box);

  /* Create Event Handlers */
  XtAddEventHandler(Box, ExposureMask, False, RefreshMe, NULL);
  XtAddEventHandler(Box, ButtonPressMask, False, ExitNow, NULL);

  /* This pixmap acts as backing store for the screen */
  pixmap = XCreatePixmap(XtDisplay(Box), 
			 RootWindowOfScreen(XtScreen(Box)), 
			 iWidth, iHeight, 
			 DefaultDepth(XtDisplay(Box), XtWindow(Box)));
  
  /* Create some gc's */
  gcDraw = XCreateGC(XtDisplay(Box), pixmap, 0, NULL);
  gcErase = XCreateGC(XtDisplay(Box), pixmap, 0, NULL);

  XSetForeground(XtDisplay(Box), gcDraw, 
		 BlackPixel(XtDisplay(Box), DefaultScreen(XtDisplay(Box))));
  XSetForeground(XtDisplay(Box), gcErase, 
		 WhitePixel(XtDisplay(Box), DefaultScreen(XtDisplay(Box))));
  XSetBackground(XtDisplay(Box), gcDraw, 
		 WhitePixel(XtDisplay(Box), DefaultScreen(XtDisplay(Box))));
  XSetBackground(XtDisplay(Box), gcErase, 
		 BlackPixel(XtDisplay(Box), DefaultScreen(XtDisplay(Box))));

  /* Erase the backing pixmap */
  XFillRectangle(XtDisplay(Box), pixmap, gcErase, 0, 0, iWidth, iHeight);

  XtRealizeWidget(Toplevel);

  /* Init Stefan's hash table */
  for (int i = 0; i < TABLE_SIZE; i++)
    bit_table[i] = NULL;
}



/************************************************************************ 
 *  FUNCTION: TickleXWindow
 *  HISTORY: 
 *     11/22/93  ESF  Created
 *  PURPOSE: 
 *     Since EThreads is a user space package, when Xt blocks, the 
 *   threads don't get to run.  This is not good, so we add a timer, 
 *   which goes off every little bit, and spends some time doing
 *   nothing, which lets the clock tick in user space and eventually
 *   this thread gets preempted, and other threads run.  This is a cheezy
 *   (but effective) way to write a multithreaded X app with user space
 *   threads.
 *  NOTES: 
 ************************************************************************/
void TickleXWindow(Widget w, XtPointer closure, XtPointer calldata)
{
  Int j, k;

#ifdef LINUX_GCC
  Int iLim = 2000;
#else
  Int iLim = 20000;
#endif

  /* Do nothing, and enjoy it */
  for(j=0, k=3; j!=iLim; j++)
    k = k*k;

  /* Rearm the timer */
  timer = XtAppAddTimeOut(context, _hack_ticks_, TickleXWindow, Box);
}


/************************************************************************ 
 *  FUNCTION: MainLoop
 *  HISTORY: 
 *     11/22/93  ESF  Created
 *  PURPOSE:
 *     This is the thread function for the X thread. 
 *  NOTES: 
 ************************************************************************/
void MainLoop(void *bob)
{
  XtAppMainLoop(context);
}


/************************************************************************ 
 *  FUNCTION: DrawBitmap
 *  HISTORY: 
 *     ??/??/93  ???  Created
 *     11/??/93   SM  Hash table speedup
 *     11/22/93  ESF  Pixmap backing store, Xt.
 *  PURPOSE: 
 *     Draws the specified bitmap at the specified coordinates with
 *   the forground and background colors set in gcDraw.
 *  NOTES: 
 ************************************************************************/
void DrawBitmap(char *data, int width, int height, int x, int y)
{
  Pixmap   pixPicture;
  bithash *curr_location;
  sigset_t sigMask = sigblock(sigmask(SIGVTALRM));
  
  /* Lookup the bitmap in the hash table */
  curr_location = hash_find_bitmap(data);

  /* If it's not in the hash table, insert it in there */
  if (curr_location==NULL)
    {
      pixPicture = XCreateBitmapFromData (XtDisplay(Box), XtWindow(Box), 
					  data, width, height);
      hash_bitmap (data, pixPicture);
    }
  else
    pixPicture = curr_location->xid;

  /* Copy the pixmap to the screen and to the backing store */
  XCopyPlane (XtDisplay(Box), pixPicture, XtWindow(Box), gcDraw, 
	      0, 0, width, height, x, y, 1);

  XCopyPlane (XtDisplay(Box), pixPicture, pixmap, gcDraw, 
	      0, 0, width, height, x, y, 1);

  /* And do it this week, please :) */
  XFlush (XtDisplay(Box));

  sigsetmask(sigMask);
}


/************************************************************************ 
 *  FUNCTION: EraseBitmap
 *  HISTORY: 
 *     ??/??/93  ???  Created
 *     11/??/93   SM  Hash table speedup
 *     11/22/93  ESF  Pixmap backing store, Xt.
 *  PURPOSE: 
 *     Draws the specified bitmap at the specified coordinates with
 *   the forground and background colors set in gcDraw.
 *  NOTES: 
 ************************************************************************/
void EraseBitmap(char *data, int width, int height, int x, int y)
{
  sigset_t sigMask = sigblock(sigmask(SIGVTALRM));

  /* Erase the rectangle that holds (held) the bitmap */
  XFillRectangle(XtDisplay(Box), pixmap, gcErase, x, y, width, height);
  XFillRectangle(XtDisplay(Box), XtWindow(Box), gcErase, x, y, width, height);
  XFlush(XtDisplay(Box));

  sigsetmask(sigMask);
}



/************************************************************************ 
 *  FUNCTION: ExitNow
 *  HISTORY: 
 *     11/22/93  ESF  Created
 *  PURPOSE: 
 *     Callback for the mouse click event.  Exits, ungracefully :)
 *  NOTES: 
 ************************************************************************/
void ExitNow(Widget w, XtPointer closure, XtPointer calldata)
{
  exit(0);
}



/************************************************************************ 
 *  FUNCTION: RefreshMe
 *  HISTORY: 
 *     11/22/93  ESF  Created
 *  PURPOSE: 
 *     Callback for Expose events.
 *  NOTES: 
 ************************************************************************/
void RefreshMe(Widget w, XtPointer closure, XtPointer calldata)
{
  XExposeEvent *event = (XExposeEvent *)calldata;
  sigset_t sigMask = sigblock(sigmask(SIGVTALRM));

  /* Redraw the exposed part from the backing store pixmap */
  XCopyArea(XtDisplay(Box), pixmap, XtWindow(Box), gcDraw, 
	    event->x, event->y, event->width, event->height, 
	    event->x, event->y);
  
  sigsetmask(sigMask);
}



/************************************************************************ 
 *  FUNCTION: EraseBitmap
 *  HISTORY: 
 *     11/??/93   SM  Created.
 *  PURPOSE: 
 *     Finds a bitmap in the hashtable.
 *  NOTES: 
 ************************************************************************/
bithash *hash_find_bitmap (char *bitmap)
{
  bithash *tmp;
  unsigned long loc = (unsigned long) bitmap % TABLE_SIZE;

  for (tmp = bit_table[loc];
       tmp != NULL && tmp->id != bitmap;
       tmp = tmp->next);

  return (tmp);
}



/************************************************************************ 
 *  FUNCTION: EraseBitmap
 *  HISTORY: 
 *     11/??/93   SM  Created
 *  PURPOSE: 
 *     Inserts a bitmap in the hash table.
 *  NOTES: 
 ************************************************************************/
void hash_bitmap (char *id, Pixmap xid)
{
  bithash *New = (bithash *) malloc(sizeof (bithash));
  unsigned long loc = (unsigned long) id % TABLE_SIZE;
  bithash *curr_location;

  if (bit_table[loc] == NULL)
    bit_table[loc] = New;
  else
    {
      for (curr_location = bit_table[loc];
	   curr_location->next != NULL;
	   curr_location = curr_location->next);
      curr_location->next = New;
    }
  New->next = NULL;
  New->id = id;
  New->xid = xid;
}

